#!/usr/bin/env python

import pickle
import sys
from lib.cluster import Machine
from copy import deepcopy
import time
from lib.instance import readInstance

# BEST_FIT
# BEST_AVAILABILITY
# WORST_FIT
# RANDOM
MACHINE_SELECTION_POLICY = "BEST_AVAILABILITY" 

def __machineSelector(machines, demand, function, policy = MACHINE_SELECTION_POLICY):
	best = None

	assert policy in ["BEST_FIT", "WORST_FIT" , "BEST_AVAILABILITY", "RANDOM"]

	if machines:
		if policy == "BEST_AVAILABILITY":
			best = max(machines, key=lambda m : m.availability )
		elif policy == "BEST_FIT":
			best = min(machines, key=lambda m : m.residualCapacity() )
		elif policy == "WORST_FIT":
			best = max(machines, key=lambda m : m.residualCapacity() )
		else:
			best = machines[0]
	
	return best

def __packRequests(instance, loadBalancing = True, sortedRequests = False, selectionPolicy = "RANDOM", factor = 1.0):
	"""Assign requests to machines starting from the most reliable one"""
	
	Machine.LOAD_BALANCING = loadBalancing
	Machine.VM_SIZE_FACTOR = float(factor)
	
	requests = instance.requests[:]
	
	if sortedRequests:
		requests.sort(key=lambda r : r.penalty)
	else:
		requests.sort(key=lambda r : r.name, reverse=True)
	
	while requests:
		r = requests.pop()
		
		m = __machineSelector([m for m in instance.machines if m.isAvailable(r.demand, r.function)],r.residual, r.function, selectionPolicy)
		if m == None:
			continue

		m.addRequest(r, min(r.residual,  m.residualCapacity(r.function)))

		if r.residual > 0:
			requests.append(r)
			
def addProtection(vm, machines):
	backup = None
	machine = __machineSelector([m for m in machines if not m.hasVMCopy(vm) and m.residualCapacity() >= vm.volume], vm.volume, vm.function)
	if machine:
		backup = vm.backup(machine)
	return backup

def __addProtection(instance):
	virtualMachines = sum([m.virtualMachines for m in instance.machines], [])
	backupFound = True
	
	while backupFound:
		backupFound = False
		for vm in virtualMachines:
			backupFound = True if addProtection(vm, instance.machines) else False
					
def solve(instance, loadBalancing = True, protection = False, sortedRequests = False, selectionPolicy = "BEST_AVAILABILITY", factor = 1.0):
	
	instance = deepcopy(instance)
	
	__packRequests(instance, loadBalancing, sortedRequests, selectionPolicy, factor)
	if protection:
		__addProtection(instance)
	
	return instance

if __name__ == '__main__':
	
	#instance = generatore.generate(1,False)
	instance = readInstance(sys.argv[1])
	
	start_time = time.time()
	
	balancing = sys.argv[2] == "True"
	protection = sys.argv[3] == "True"
	instance.objective = sys.argv[4]

	instance = solve(instance, balancing, protection)

	end_time = time.time()

	print(instance.stats())
	print("TotelTime %.6f" % (end_time - start_time))
	if len(sys.argv) == 6:
		instance.savePlacement(sys.argv[5], "TEXT")
