#!/usr/bin/env python

class Request(object):
    
    __counter = 0
    
    def __init__(self, entryPoints = [], function = None, demand = 0, penalty = 1, objid = None):
        if objid:
            assert Request.__counter <= int(objid)
            Request.__counter = int(objid)
        else:
            Request.__counter += 1
        self.__name = "r"+str(Request.__counter)
        self.__demand = float(demand)
        self.__served = 0.0
        self.__penalty = float(penalty)
        self.__function = function
        self.__entryPoints = entryPoints
        self.__masterVMs = {}
        
    @property
    def name(self):
        return self.__name
    
    @property
    def penalty(self):
        return self.__penalty
    
    @property
    def demand(self):
        return self.__demand
    
    @property
    def served(self):
        return self.__served
    
    @property
    def residual(self):
        assert(self.__demand >= self.__served)
        return self.__demand - self.__served
    
    def serve(self, servedDemand, virtualMachine):
        assert(servedDemand <= self.residual)
        self.__served += float(servedDemand)
        assert(virtualMachine.isMaster())
        self.__masterVMs.setdefault(virtualMachine, 0.0)
        self.__masterVMs[virtualMachine] += servedDemand
 
    def restore(self, restoredDemand, virtualMachine):
        assert(restoredDemand <= self.__served)
        assert(virtualMachine in self.__masterVMs)
        self.__served -= float(restoredDemand)
        
        if restoredDemand == self.__masterVMs[virtualMachine]:
            self.__masterVMs.pop(virtualMachine)
        else:
            self.__masterVMs[virtualMachine] -= float(restoredDemand)
        
        assert(self.__served >= 0)
    
    @property
    def function(self):
        return self.__function
    
    def isAssigned(self):
        return self.residual == 0
    
    @property
    def masters(self):
        return self.__masterVMs.keys()
    
    @property
    def entryPoints(self):
        return self.__entryPoints
    
    def __str__(self):
        return "<"+self.__name+", f:"+str(self.__function)+", ep:"+str(self.__entryPoints)+", d:"+str(self.__demand) +">"
    
    def __repr__(self):
        return "<"+self.__name+">"
    
    def __deepcopy__(self, memo):
        cls = self.__class__
        newone = cls.__new__(cls)
        
        newone.__name = self.__name
        newone.__demand = self.__demand
        newone.__served = 0.0
        newone.__penalty = self.__penalty
        newone.__function = self.__function
        newone.__entryPoints = self.__entryPoints
        newone.__masterVMs = {}
        
        return newone
    
    
if __name__ == '__main__':
    requests = [Request() for _ in range(0,3)]
    print (requests)
